@@ -18,6 +18,7 @@
 static bool PGrnWALEnabled = false;
 static size_t PGrnWALMaxSize = 0;
 static bool PGrnWALResourceManagerEnabled = false;
+static size_t PGrnWALMaxBulkInsertRecordSize = 16 * 1024 * 1024; /* 16MiB */
 
 bool
 PGrnWALGetEnabled(void)
@@ -67,6 +68,24 @@ PGrnWALResourceManagerDisable(void)
 	PGrnWALResourceManagerEnabled = false;
 }
 
+bool
+PGrnWALResourceManagerIsOnlyEnabled(void)
+{
+	return !PGrnWALEnabled && PGrnWALResourceManagerEnabled;
+}
+
+size_t
+PGrnWALGetMaxBulkInsertRecordSize(void)
+{
+	return PGrnWALMaxBulkInsertRecordSize;
+}
+
+void
+PGrnWALSetMaxBulkInsertRecordSize(size_t size)
+{
+	PGrnWALMaxBulkInsertRecordSize = size;
+}
+
 static bool
 PGrnWALAnyEnabled(void)
 {
@@ -137,6 +156,7 @@ struct PGrnWALData_
 {
 	Relation index;
 	grn_obj *table;
+	bool bulkInserting;
 #ifdef PGRN_SUPPORT_WAL
 	GenericXLogState *state;
 	unsigned int nUsedPages;
@@ -514,6 +534,7 @@ PGrnWALStart(Relation index)
 
 	data = palloc(sizeof(PGrnWALData));
 	data->table = NULL;
+	data->bulkInserting = false;
 
 #	ifdef PGRN_SUPPORT_WAL
 	if (PGrnWALEnabled)
@@ -587,6 +608,68 @@ PGrnWALAbort(PGrnWALData *data)
 #endif
 }
 
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+static void
+PGrnWALBulkInsertStartCustom(PGrnWALData *data, grn_obj *table)
+{
+	if (!PGrnWALResourceManagerEnabled)
+		return;
+
+	{
+		grn_obj *buffer = &(buffers->walRecord);
+		PGrnWALRecordCommon record = {
+			.dbID = MyDatabaseId,
+			.dbEncoding = GetDatabaseEncoding(),
+			.dbTableSpaceID = MyDatabaseTableSpace,
+		};
+
+		GRN_BULK_REWIND(buffer);
+		PGrnWALRecordBulkInsertWriteStart(buffer, &record, table);
+	}
+}
+#endif
+
+void
+PGrnWALBulkInsertStart(PGrnWALData *data, grn_obj *table)
+{
+	if (!data)
+		return;
+	if (grn_obj_is_temporary(ctx, table))
+		return;
+
+	data->table = table;
+	data->bulkInserting = true;
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+	PGrnWALBulkInsertStartCustom(data, table);
+#endif
+}
+
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+static void
+PGrnWALBulkInsertFinishCustom(PGrnWALData *data)
+{
+	if (!PGrnWALResourceManagerEnabled)
+		return;
+
+	{
+		grn_obj *buffer = &(buffers->walRecord);
+		PGrnWALRecordBulkInsertWriteFinish(buffer);
+	}
+}
+#endif
+
+void
+PGrnWALBulkInsertFinish(PGrnWALData *data)
+{
+	if (!data)
+		return;
+
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+	PGrnWALBulkInsertFinishCustom(data);
+#endif
+	data->bulkInserting = false;
+}
+
 #ifdef PGRN_SUPPORT_WAL
 static void
 PGrnWALInsertStartGeneric(PGrnWALData *data, grn_obj *table, size_t nColumns)
@@ -625,15 +708,29 @@ PGrnWALInsertStartCustom(PGrnWALData *data, grn_obj *table, size_t nColumns)
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
+		grn_obj *buffer = &(buffers->walRecord);
 		PGrnWALRecordCommon record = {
 			.dbID = MyDatabaseId,
 			.dbEncoding = GetDatabaseEncoding(),
 			.dbTableSpaceID = MyDatabaseTableSpace,
 		};
 
-		GRN_BULK_REWIND(buffer);
-		PGrnWALRecordInsertWriteStart(buffer, &record, table, nColumns);
+		if (data->bulkInserting)
+		{
+			if (PGrnWALMaxBulkInsertRecordSize > 0 &&
+				GRN_TEXT_LEN(buffer) > PGrnWALMaxBulkInsertRecordSize)
+			{
+				PGrnWALRecordBulkInsertWriteFinish(buffer);
+				GRN_BULK_REWIND(buffer);
+				PGrnWALRecordBulkInsertWriteStart(buffer, &record, table);
+			}
+			PGrnWALRecordBulkInsertWriteRecordStart(buffer, &record, nColumns);
+		}
+		else
+		{
+			GRN_BULK_REWIND(buffer);
+			PGrnWALRecordInsertWriteStart(buffer, &record, table, nColumns);
+		}
 	}
 }
 #endif
@@ -670,8 +767,11 @@ PGrnWALInsertFinishCustom(PGrnWALData *data)
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteFinish(buffer);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+			PGrnWALRecordBulkInsertWriteRecordFinish(buffer);
+		else
+			PGrnWALRecordInsertWriteFinish(buffer);
 	}
 }
 #endif
@@ -718,8 +818,11 @@ PGrnWALInsertColumnStartCustom(PGrnWALData *data,
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteColumnStart(buffer, name, nameSize);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+			PGrnWALRecordBulkInsertWriteColumnStart(buffer, name, nameSize);
+		else
+			PGrnWALRecordInsertWriteColumnStart(buffer, name, nameSize);
 	}
 }
 #endif
@@ -772,8 +875,11 @@ PGrnWALInsertColumnValueKeyCustom(PGrnWALData *data,
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteColumnValueKey(buffer, key, keySize);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+			PGrnWALRecordBulkInsertWriteColumnValueKey(buffer, key, keySize);
+		else
+			PGrnWALRecordInsertWriteColumnValueKey(buffer, key, keySize);
 	}
 }
 #endif
@@ -906,8 +1012,17 @@ PGrnWALInsertColumnValueBulkCustom(PGrnWALData *data,
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteColumnValueBulk(buffer, name, nameSize, value);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+		{
+			PGrnWALRecordBulkInsertWriteColumnValueBulk(
+				buffer, name, nameSize, value);
+		}
+		else
+		{
+			PGrnWALRecordInsertWriteColumnValueBulk(
+				buffer, name, nameSize, value);
+		}
 	}
 }
 #endif
@@ -968,9 +1083,17 @@ PGrnWALInsertColumnValueVectorCustom(PGrnWALData *data,
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteColumnValueVector(
-			buffer, name, nameSize, value);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+		{
+			PGrnWALRecordBulkInsertWriteColumnValueVector(
+				buffer, name, nameSize, value);
+		}
+		else
+		{
+			PGrnWALRecordInsertWriteColumnValueVector(
+				buffer, name, nameSize, value);
+		}
 	}
 }
 #endif
@@ -1032,9 +1155,17 @@ PGrnWALInsertColumnUValueVectorCustom(PGrnWALData *data,
 		return;
 
 	{
-		grn_obj *buffer = &(buffers->walBuffer);
-		PGrnWALRecordInsertWriteColumnValueUVector(
-			buffer, name, nameSize, value);
+		grn_obj *buffer = &(buffers->walRecord);
+		if (data->bulkInserting)
+		{
+			PGrnWALRecordBulkInsertWriteColumnValueUVector(
+				buffer, name, nameSize, value);
+		}
+		else
+		{
+			PGrnWALRecordInsertWriteColumnValueUVector(
+				buffer, name, nameSize, value);
+		}
 	}
 }
 #endif
@@ -1386,7 +1517,10 @@ PGrnWALSetSourceIDsCustom(Relation index, grn_obj *column, grn_obj *sourceIDs)
 								MyDatabaseTableSpace,
 								column,
 								sourceIDs);
-	PGrnWALRecordSetSourcesWrite(&record);
+	GRN_BULK_REWIND(&(buffers->sourceNames));
+	GRN_BULK_REWIND(&(buffers->sourceNameSizes));
+	PGrnWALRecordSetSourcesWrite(
+		&record, &(buffers->sourceNames), &(buffers->sourceNameSizes));
 }
 #endif
 
